/*
 * Copyright (C) 2002 Nikos Mavroyanopoulos
 *
 * This file is part of BOA webserver.
 *
 * Hydra is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Hydra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include "boa.h"
#include <signal.h>             /* signal */

/* Hydra Internally handled CGIs (HIC HIC)
 *
 * How does it work?
 *
 * A special thread (maybe more than one), is used for CGIs. This tread
 * uses one pipe to communicate with the others via hic_send_command().
 *
 * If a request for a HIC file arrives, the server goes as a normal CGI 
 * creates a pipe, but does not fork. It enqueues a hic_request
 * for this (HIC) thread, and sends a SIGUSR2 signal. The HIC thread is then
 * waken up, and handles all pending requests.
 *
 * The HIC thread, when receives the message, parses the CGIs,
 * one CGI at the time, and sends the output to fd.
 *
 * The benefit is that in no case the server is blocked. We may have
 * a performance problem, if the server only uses parsed CGIs.
 *
 */

#ifdef ENABLE_HIC

#include "hic.h"

static int process_hic_request(hic_params*, hic_request * cmd);
void hic_dequeue(hic_request ** head, hic_request * req);
void hic_enqueue(hic_request ** head, hic_request * req);

void *hic_main_loop(void *_hic_params)
{
   hic_request* current, *next;
   hic_params *params = _hic_params;

   while (1) {
      for (current = params->request_head;current;current=next) {

#ifdef ENABLE_SMP
         pthread_mutex_lock(&params->lock);
#endif
         next = current->next;
#ifdef ENABLE_SMP
         pthread_mutex_unlock(&params->lock);
#endif

	 if (process_hic_request( params, current) == -1) {
	    log_error_time();
	    fprintf(stderr, "hic: Error while processing command.\n");
	    exit(1);
	 }
      }
      
      /* If somebody sends a command, then he should
       * send a sigusr2 as well.
       */
      unblock_sigusr2();
      if (!params->request_head) pause();
      block_sigusr2();

      if (params->sigchld_flag)
         sigchld_run();
      if (params->sighup_flag)
         sighup_run();
      

   }

}

static int process_hic_request(hic_params* params, hic_request * req)
{
   hic_module_st *hst;
   hic_stuff* cmd = &req->command;
   
#ifdef ENABLE_SMP
   pthread_mutex_lock(&params->lock);
#endif
   hic_dequeue( &params->request_head, req);
#ifdef ENABLE_SMP
   pthread_mutex_unlock(&params->lock);
#endif

   hst = find_hic_appr_module(get_mime_type(cmd->path_translated), 0);
   if (hst == NULL) {
      log_error_time();
      fprintf(stderr, "Could not find HIC handler for '%s'\n",
	      cmd->path_translated);
      close(cmd->out_fd);
      free( req);
      return 0;
   }

   hst->request(cmd);
   close(cmd->out_fd);

/*
   log_error_time();
   fprintf(stderr, "HIC status: %d wrote: %d\n", cmd->status, cmd->bytes_sent);
*/

   free( req);

   return 0;
}

#ifdef ENABLE_SMP
/* We use this index, to send messages to different
 * HIC thread every time.
 */
static pthread_mutex_t hic_index_lock = PTHREAD_MUTEX_INITIALIZER;
static int hic_index;
#endif

/* Sends a command to a HIC thread.
 */
int hic_send_command(request * req, int out_fd)
{
   hic_request *x;
   hic_params *params;

   if (global_hic_params_size == 0) /* hic is disabled */
      return -1;

   /* choose a random hic thread */
   if (global_hic_params_size > 1) {
#ifdef ENABLE_SMP
      pthread_mutex_lock( &hic_index_lock);
#endif
      params = &global_hic_params[ hic_index];
      hic_index++;
      hic_index %= global_hic_params_size;
#ifdef ENABLE_SMP
      pthread_mutex_unlock( &hic_index_lock);
#endif
   } else {
      params = &global_hic_params[0];
   }

   x = malloc( sizeof( hic_request));
   if (x==NULL)
      return -1;

   x->next = x->prev = NULL;
   x->command.cgi_env = req->cgi_env;
   x->command.http_version = req->http_version;
   x->command.post_data_fd = req->post_data_fd.fds[0];
   x->command.out_fd = out_fd;
   x->command.path_translated = req->path_translated;

#ifdef ENABLE_SMP
   pthread_mutex_lock(&params->lock);
#endif
   
   hic_enqueue( &params->request_head, x);
#ifdef ENABLE_SMP
   /* Notify the HIC thread about the pending request 
    */
   pthread_kill( params->tid, SIGUSR2);
#endif

#ifdef ENABLE_SMP
   pthread_mutex_unlock(&params->lock);
#endif

   return 1;

}

/* Implement a basic queue for HIC stuff. This is created after
 * queue.c.
 */

#include "queue.h" 

ENQUEUE_FUNCTION( hic_enqueue, hic_request)
DEQUEUE_FUNCTION( hic_dequeue, hic_request)

#endif	/* ENABLE_HIC */
