/*
 *  Hydra, an http server
 *  Copyright (C) 2002 Nikos Mavroyanopoulos <nmav@gnutls.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 1, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/* $Id: hic_modules.c,v 1.10 2003/02/18 18:35:27 nmav Exp $ */

/* This file includes support for dynamically loaded HIC modules
 * All modules added to module_table[] will be dlopen()ed at startup.
 *
 * Also hic symbols will be resolved.
 */

#include "boa.h"

#ifdef ENABLE_HIC
# include <dlfcn.h>
#endif

/* A dynamic module, must provide three functions:
 * void     _php_hic_shutdown(void);
 * void     _php_hic_init(void);
 * off_t    _php_hic_request(hic_stuff *hc);
 *
 * This is from php.
 */


static hic_module_st* module_hashtable[MODULE_HASHTABLE_SIZE];

/*
 * Name: add_module
 *
 * Description: add an module entry
 */

void add_hic_module(const char *module, const char* sym_prefix, const char* content_type)
{
#ifndef ENABLE_HIC
    fprintf(stderr, "Cannot open '%s' since dynamic module opening, is not supported, "
     "or HIC has been disabled.\n", module);
#else
char symbol[128];
int hash;
hic_module_st* old, *start;

    int sym_prefix_len;
    void * handle;

    /* sanity checking */
    if (module == NULL || sym_prefix == NULL || content_type == NULL) {
        DIE("NULL values sent to add_module");
    }

    hash = get_hic_module_hash_value( content_type);
    start = old = module_hashtable[hash];
    
    if ( old != NULL) {
       /* find next empty */
       do {
          hash = (hash + 1) % MODULE_HASHTABLE_SIZE;
          
          old = module_hashtable[hash];

          if (start == old) {
             DIE("Module hashtable is full.");
          }
          
       } while( old != NULL);
    }
    
    /* old was found, and is empty. */
    
    old = malloc( sizeof(hic_module_st));
    if (old==NULL) {
       DIE("malloc() failed.");
    }

    old->sym_prefix = strdup( sym_prefix);
    if (old->sym_prefix == NULL) {
       DIE("strdup() failed.");
    }

    old->content_type = strdup( content_type);
    if (old->content_type == NULL) {
       DIE("strdup() failed.");
    }

    old->content_type_len = strlen( content_type);


/* Handle OSs that only implement RTLD_LAZY. */
#ifndef RTLD_NOW
#ifdef __GNUC__
#warning "OS does not implement RTLD_NOW, using RTLD_LAZY; it's probably safe."
#endif /* __GNUC__ */
#define RTLD_NOW RTLD_LAZY
#endif /* not RTLD_NOW */
    handle = dlopen( module, RTLD_NOW);
    if ( handle==NULL) {
       fprintf(stderr, "Could not load module '%s'. Dlopen failed: %s\n", module, dlerror());
       exit(1);
    }
    old->dl_handle = handle;


    sym_prefix_len = strlen( sym_prefix);
    if (sym_prefix_len + 20 > sizeof(symbol)) {
        DIE("Symbol prefix is too long.");
    }

    /* Resolve shutdown */
    strcpy( symbol, sym_prefix);
    strcat( symbol, "_hic_shutdown");
   
    old->shutdown = dlsym( handle, symbol);
    if ( old->shutdown == NULL) {
       fprintf(stderr, "Could not resolve %s_hic_shutdown symbol", sym_prefix);
       exit(1);
    }

    /* Resolve init */
    strcpy( symbol, sym_prefix);
    strcat( symbol, "_hic_init");

    old->init = dlsym( handle, symbol);
    if ( old->init == NULL) {
       fprintf(stderr, "Could not resolve %s_hic_init symbol", sym_prefix);
       exit(1);
    }
    old->init(); /* Run the initialization stuff */

    /* Resolve request */
    strcpy( symbol, sym_prefix);
    strcat( symbol, "_hic_request");

    old->request = dlsym( handle, symbol);
    if ( old->request == NULL) {
       fprintf(stderr, "Could not resolve %s_hic_request symbol", sym_prefix);
       exit(1);
    }

    module_hashtable[hash] = old;

    return;
#endif
}


/* add_hic_action
 *
 * Like add_hic_module() but associates the file type with a
 * specific action (executable to run with)
 */

void add_hic_action(const char *action, const char* file_type)
{
int hash;
hic_module_st* old, *start;

    /* sanity checking */
    if (action == NULL || file_type == NULL) {
        DIE("NULL values sent to add_hic_action");
    }

    hash = get_hic_module_hash_value( file_type);
    start = old = module_hashtable[hash];
    
    if ( old != NULL) {
       /* find next empty */
       do {
          hash = (hash + 1) % MODULE_HASHTABLE_SIZE;
          
          old = module_hashtable[hash];

          if (start == old) {
             DIE("Module hashtable is full.");
          }
          
       } while( old != NULL);
    }
    
    /* old was found, and is empty. */
    
    old = malloc( sizeof(hic_module_st));
    if (old==NULL) {
       DIE("malloc() failed.");
    }

    old->sym_prefix = NULL;
 
    old->content_type = strdup( file_type);
    if (old->content_type == NULL) {
       DIE("strdup() failed.");
    }

    old->content_type_len = strlen( file_type);

    old->action = strdup( action);
    if ( old->action == NULL) {
       DIE("strdup() failed.");    
    }

    old->dl_handle = NULL;
    old->shutdown = NULL;
    old->init = NULL;
    old->request = NULL;

    module_hashtable[hash] = old;

    return;
}

/*
 * Name: find_hic_appr_module
 *
 * Description: Locates the appropriate HIC module for the given file.
 * Actually ones needs this to get the dlsymed() functions.
 * 
 * Returns:
 *
 * a pointer to a hic_module_st structure or NULL if not found
 */

hic_module_st *find_hic_appr_module(const char *content_type, int content_type_len)
{
int i, hash;

   if (content_type == NULL) return NULL;
   if (content_type_len == 0) content_type_len = strlen( content_type);

   hash = get_hic_module_hash_value( content_type);
   for (i=hash;i<MODULE_HASHTABLE_SIZE;i++) {
      if (module_hashtable[i] == NULL) break;

      if ( content_type_len != module_hashtable[i]->content_type_len) continue;

      if (memcmp( content_type, module_hashtable[i]->content_type, 
      	content_type_len) == 0) {
        /* FOUND! */
      	return module_hashtable[i];
      }
   }
 
   return NULL;

}


/*
 * Empties the hic modules table, deallocating any allocated memory.
 */

void dump_hic_modules(void)
{
int i;

    for (i = 0; i < MODULE_HASHTABLE_SIZE; ++i) { /* these limits OK? */
        if (!module_hashtable[i]) continue;

#ifdef ENABLE_HIC
        free( module_hashtable[i]->sym_prefix);
        free( module_hashtable[i]->content_type);

        if (module_hashtable[i]->shutdown)
           module_hashtable[i]->shutdown(); /* Run the deinitialization stuff */

        if (module_hashtable[i]->dl_handle)
           dlclose( module_hashtable[i]->dl_handle);
#endif

        free( module_hashtable[i]->action);

        free( module_hashtable[i]);
        module_hashtable[i] = NULL;
    }
}
